#------------------------------------------------------------------------------
#    plSokoban Suite - perl OOP sokoban implementation
#    Copyright (c) 2002 Arne Bochem <webmaster@arneb.de>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#------------------------------------------------------------------------------

use strict;
use Sokoban::Element::Wall;
use Sokoban::Element::Empty;
use Sokoban::Element::Player;
use Sokoban::Element::Boulder;

package Sokoban::Engine;

sub new
{
	my $class = shift;
	my $field = shift;
	my $player;
	my $boulders;
	my $x;
	my $y;

	my $self = bless({
	              '_player' => $player,
	              '_field' => $field,
	              '_boulders' => $boulders,
	              '_width' => $x,
	              '_height' => $y
	             }, $class);

	for ($y = 0; $y < scalar(@{$field}); $y++)
	{
		for ($x = 0; $x < scalar(@{$field->[$y]}); $x++)
		{
			$field->[$y][$x] = create_element($self, $field->[$y][$x], $x, $y);
			$field->[$y][$x]->player() and ($player = $field->[$y][$x]);
			$field->[$y][$x]->boulder() and $boulders++;
		}
	}

	$field or die('No field created.');
	$player or die('No player found.');

	$self->{'_player'} = $player;	
	$self->{'_field'} = $field;	
	$self->{'_boulders'} = $boulders;
	$self->{'_width'} = $x;	
	$self->{'_height'} = $y;	

	return $self;
}

sub create_element
{
	my ($parent, $element_type, $x, $y) = @_;

	my %element_types = (
	                     '#' => sub { return Sokoban::Element::Wall->new(@_, 0); },
	                     ' ' => sub { return Sokoban::Element::Empty->new(@_, 0); },
	                     '' => sub { return Sokoban::Element::Empty->new(@_, 0); },
	                     'o' => sub { return Sokoban::Element::Empty->new(@_, 1); },
	                     '@' => sub { return Sokoban::Element::Player->new(@_, 0); },
	                     'x' => sub { return Sokoban::Element::Boulder->new(@_, 0); }
	                    );

	return $element_types{$element_type}->($parent, $x, $y, $element_type);
}

sub redefine
{
	my ($self, $element, $element_type) = @_;

	my %element_types = (
	                     '#' => sub { return Sokoban::Element::Wall->new(@_); },
	                     ' ' => sub { return Sokoban::Element::Empty->new(@_); },
	                     '@' => sub { return Sokoban::Element::Player->new(@_); },
	                     'x' => sub { return Sokoban::Element::Boulder->new(@_); }
	                    );
	
	my $y = $element->{'_y'};
	my $x = $element->{'_x'};

	$self->{'_field'}[$y][$x] = $element_types{$element_type}->($element->{'_parent'}, $x, $y, $element_type, $element->{'_target'});
	($element_type eq '@') and ($self->{'_player'} = $self->{'_field'}[$y][$x]);
}

sub get_element
{
	my ($self, $x, $y) = @_;

	(($x < $self->{'_width'}) and ($y < $self->{'_height'}) and ($x >= 0) and ($y >= 0)) or return(0);
	return $self->{'_field'}[$y][$x];
}

sub boulder_hit
{
	my $self = shift;
	$self->{'_boulders'}--;
}

sub boulder_remove
{
	my $self = shift;
	$self->{'_boulders'}++;
}

sub boulders_done
{
	my $self = shift;
	return !$self->{'_boulders'};
}

sub set_player
{
	my ($self, $new_player) = @_;

	$self->{'_player'} = $new_player;
}

sub command
{
	my ($self, $direction) = @_;

	$self->{'_player'}->move($direction);
}

sub printable
{
	my $self = shift;
	my $field = [];

	for (my $y = 0; $y < $self->{'_height'}; $y++)
	{
		my $line = [];
		for (my $x = 0; $x < $self->{'_width'}; $x++)
		{
			push @{$line}, $self->{'_field'}[$y][$x]->element_type();
		}
		push @{$field}, $line;
	}
	
	return $field;
}

sub ascii_art
{
	my $self = shift;
	my $field = $self->printable();

	my $output = join("\n", map { join('', @{$_}); } @{$field});
	return $output."\n";
}

1;
